package com.gemserk.commons.animation.handlers;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import com.gemserk.commons.animation.Animation;
import com.gemserk.commons.animation.MockAnimation;
import com.gemserk.commons.animation.timeline.LinearInterpolatorFactory;
import com.gemserk.commons.animation.timeline.TimelineAnimation;
import com.gemserk.commons.animation.timeline.TimelineBuilder;
import com.gemserk.commons.animation.timeline.TimelineValueBuilder;

public class AnimationHandlerManagerTest {

	@Test
	public void onEndShouldNotBeCalledIfAnimationNotFinished() {
		Animation animation = new MockAnimation() {
			{
				setFinished(false);
				setStarted(false);
			}
		};

		MockAnimationEventHandler animationEventHandler = new MockAnimationEventHandler();

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", animationEventHandler);
		animationHandlerManager.checkAnimationChanges();

		assertEquals(false, animationEventHandler.onFinishCalled);
	}
	
	@Test
	public void onEndShouldBeCalledOnceIfAnimationFinishedAndNeverAgainIfAnimationNotStartedAgain() {
		MockAnimation animation = new MockAnimation() {
			{
				setStarted(false);
				setFinished(false);
			}
		};

		MockAnimationEventHandler animationEventHandler = new MockAnimationEventHandler();

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", animationEventHandler);
		
		animationHandlerManager.checkAnimationChanges();
		assertEquals(false, animationEventHandler.onStartCalled);
		assertEquals(false, animationEventHandler.onFinishCalled);
		
		animation.setStarted(true);
		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();
		assertEquals(true, animationEventHandler.onStartCalled);
		assertEquals(false, animationEventHandler.onFinishCalled);

		animation.setFinished(true);
		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();
		assertEquals(false, animationEventHandler.onStartCalled);
		assertEquals(true, animationEventHandler.onFinishCalled);

		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();
		assertEquals(false, animationEventHandler.onFinishCalled);
		assertEquals(false, animationEventHandler.onStartCalled);
		
		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();
		assertEquals(false, animationEventHandler.onFinishCalled);
		assertEquals(false, animationEventHandler.onStartCalled);
	}

	@Test
	public void onFinishShouldBeCalledOnlyOnceIfAnimationFinished() {
		
		MockAnimation animation = new MockAnimation() {
			{
				setStarted(true);
				setFinished(false);
			}
		};

		MockAnimationEventHandler animationEventHandler = new MockAnimationEventHandler();

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", animationEventHandler);
		animationHandlerManager.checkAnimationChanges();
		
		assertEquals(false, animationEventHandler.onFinishCalled);
		
		animation.setFinished(true);
		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();

		assertEquals(true, animationEventHandler.onFinishCalled);
		
		animationEventHandler.reset();
		animationHandlerManager.checkAnimationChanges();

		assertEquals(false, animationEventHandler.onFinishCalled);
	}

	@Test
	public void onStartShouldBeCalledIfAnimationStartedOnlyOnce() {
		MockAnimation animation = new MockAnimation() {
			{
				setStarted(true);
			}
		};

		MockAnimationEventHandler animationEventHandler = new MockAnimationEventHandler();

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", animationEventHandler);
		animationHandlerManager.checkAnimationChanges();

		assertEquals(true, animationEventHandler.onStartCalled);
		animationEventHandler.onStartCalled = false;

		animationHandlerManager.checkAnimationChanges();

		assertEquals(false, animationEventHandler.onStartCalled);
	}

	@Test
	public void onStartShouldBeCalledTwiceIfAnimationReseted() {
		MockAnimation animation = new MockAnimation() {
			{
				setStarted(true);
			}
		};

		MockAnimationEventHandler animationEventHandler = new MockAnimationEventHandler();

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", animationEventHandler);
		animationHandlerManager.checkAnimationChanges();

		assertEquals(true, animationEventHandler.onStartCalled);
		animationEventHandler.onStartCalled = false;

		animationHandlerManager.checkAnimationChanges();

		assertEquals(false, animationEventHandler.onStartCalled);

		animation.setStarted(false);
		animationHandlerManager.checkAnimationChanges();
		animation.setStarted(true);
		animationEventHandler.onStartCalled = false;
		animationHandlerManager.checkAnimationChanges();
		assertEquals(true, animationEventHandler.onStartCalled);
	}

	@Test
	public void testAnimationHandlerManagerWithTimelineAnimation() {

		Animation animation = new TimelineAnimation(new TimelineBuilder() {
			{
				value("myvalue", new TimelineValueBuilder<Float>() {
					{
						interpolator(LinearInterpolatorFactory.linearInterpolatorFloat());
						keyFrame(0, 100f);
						keyFrame(100, 200f);
					}
				});
			}
		}.build());

		AnimationHandlerManager animationHandlerManager = new AnimationHandlerManager();
		animationHandlerManager.handleAnimationChanges("ANIMATION", animation);
		animationHandlerManager.addAnimationHandler("ANIMATION", new AnimationEventHandler() {
			@Override
			public void onAnimationStarted(Animation animation) {
				System.out.println("animation playing!!");
			}

			@Override
			public void onAnimationFinished(Animation animation) {
				System.out.println("animation stopped!!");
			}
		});
		animationHandlerManager.checkAnimationChanges();
		animation.play();
		animationHandlerManager.checkAnimationChanges();

		animation.update(500);
		animationHandlerManager.checkAnimationChanges();
		
		animation.update(500);
		animationHandlerManager.checkAnimationChanges();
		
		animation.update(500);
		animationHandlerManager.checkAnimationChanges();
		
		animation.update(500);
		animationHandlerManager.checkAnimationChanges();
	}
	
}
